from dataclasses import dataclass
from dbt.adapters.base.relation import BaseRelation, Policy
from typing import Type, TypeVar, Optional
from hologram.helpers import StrEnum


class RelationType(StrEnum):
    Table = "table"
    View = "view"
    CTE = "cte"
    MaterializedView = "materializedview"
    External = "external"


class ProtocolVersionType(StrEnum):
    V1 = "v1"
    V2 = "v2"
    V3 = "v3"


@dataclass
class ExasolQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False


Self = TypeVar("Self", bound="BaseRelation")


@dataclass(frozen=True, eq=False, repr=False)
class ExasolRelation(BaseRelation):
    quote_policy: ExasolQuotePolicy = ExasolQuotePolicy()

    @classmethod
    def create(
        cls: Type[Self],
        database: Optional[str] = None,
        schema: Optional[str] = None,
        identifier: Optional[str] = None,
        type: Optional[RelationType] = None,
        quote_policy: Type[ExasolQuotePolicy] = quote_policy,
        **kwargs,
    ) -> Self:
        kwargs.update(
            {
                "path": {
                    "schema": schema,
                    "identifier": identifier,
                },
                "type": type,
            }
        )
        return cls.from_dict(kwargs)

    @staticmethod
    def add_ephemeral_prefix(name: str):
        return f"dbt__CTE__{name}"



import os

import pytest

pytest_plugins = ["dbt.tests.fixtures.project"]


@pytest.fixture(scope="class")
def dbt_profile_target():
    return {
        "type": "exasol",
        "threads": 1,
        "dsn": os.getenv("DBT_DSN"),
        "user": os.getenv("DBT_USER"),
        "pass": os.getenv("DBT_PASS"),
        "dbname": "DB",
    }



import pytest
from dbt.tests.adapter.basic.files import generic_test_seed_yml
from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod
from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_incremental import BaseIncremental
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import (
    BaseSingularTestsEphemeral,
)
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp

from files import (
    seeds_added_csv,
    seeds_ansits_csv,
    seeds_base_csv,
    seeds_newcolumns_csv,
)


class TestSimpleMaterializationsMyAdapter(BaseSimpleMaterializations):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_ansits_csv,
        }


class TestSingularTestsMyAdapter(BaseSingularTests):
    pass


class TestSingularTestsEphemeralMyAdapter(BaseSingularTestsEphemeral):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
        }


class TestEmptyMyAdapter(BaseEmpty):
    pass


class TestEphemeralMyAdapter(BaseEphemeral):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
        }


class TestIncrementalMyAdapter(BaseIncremental):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"base.csv": seeds_base_csv, "added.csv": seeds_added_csv}


class TestGenericTestsMyAdapter(BaseGenericTests):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
            "schema.yml": generic_test_seed_yml,
        }


class TestSnapshotCheckColsMyAdapter(BaseSnapshotCheckCols):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
            "added.csv": seeds_added_csv,
        }


class TestSnapshotTimestampMyAdapter(BaseSnapshotTimestamp):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
            "newcolumns.csv": seeds_newcolumns_csv,
            "added.csv": seeds_added_csv,
        }


class TestBaseAdapterMethod(BaseAdapterMethod):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbt',
 'dbt.adapters',
 'dbt.adapters.exasol',
 'dbt.include',
 'dbt.include.exasol',
 'tests',
 'tests.functional']

package_data = \
{'': ['*'],
 'dbt.include.exasol': ['macros/*',
                        'macros/materializations/*',
                        'macros/utils/*']}

install_requires = \
['dbt-core==1.2.1',
 'dbt-tests-adapter>=1.2.1,<2.0.0',
 'pyexasol>=0.25.0,<0.26.0']

setup_kwargs = {
    'name': 'dbt-exasol',
    'version': '1.2.1',
    'description': 'Adapter to dbt-core for warehouse Exasol',
    'long_description': '<p align="center">\n  <a href="https://github.com/tglunde/dbt-exasol/actions/workflows/main.yml">\n    <img src="https://github.com/tglunde/dbt-exasol/actions/workflows/main.yml/badge.svg?event=push" alt="Unit Tests Badge"/>\n  </a>\n  <a href="https://github.com/tglunde/dbt-exasol/actions/workflows/integration.yml">\n    <img src="https://github.com/tglunde/dbt-exasol/actions/workflows/integration.yml/badge.svg?event=push" alt="Integration Tests Badge"/>\n  </a>\n</p>\n\n# dbt-exasol\n**[dbt](https://www.getdbt.com/)** enables data analysts and engineers to transform their data using the same practices that software engineers use to build applications.\n\nPlease see the dbt documentation on **[Exasol setup](https://docs.getdbt.com/reference/warehouse-setups/exasol-setup)** for more information on how to start using the Exasol adapter.\n\n# Known isues\n## Timestamp compatibility\nCurrently we did not see any possible way to make Exasol accept the timestamp format ```1981-05-20T06:46:51``` with a \'T\' as separator between date and time part. To pass the adapter tests we had to change the timestamps in the seed files to have a <space> character instead of the \'T\' (```1981-05-20 06:46:51```).\n## Default case of identifiers\nBy default Exasol identifiers are upper case. In order to use e.g. seeds or column name aliases - you need to use upper case column names. Alternatively one could add column_quote feature in order to have all columns case sensitive.\nWe changed the seeds_base in [files.py](tests/functional/files.py) and [fixtures.py](tests/functional/fixtures.py) in order to reflect this and successfully run the adapter test cases.\nAlso, this issue leads to problems for the [base tests for docs generation](https://github.com/dbt-labs/dbt-core/blob/8145eed603266951ce35858f7eef3836012090bd/tests/adapter/dbt/tests/adapter/basic/test_docs_generate.py), since the expected model is being checked case sensitive and fails therefor for Exasol. This will be the last task in [Issue #24](https://github.com/tglunde/dbt-exasol/issues/24) regarding dbt release 1.2. We will suggest a case insensitive version of this standard test or implement for a following minor release of dbt-exasol.\n\n## Utilities shim package\nIn order to support packages like dbt-utils and dbt-audit-helper, we needed to create the [shim package exasol-utils](https://github.com/tglunde/exasol-utils). In this shim package we need to adapt to parts of the SQL functionality that is not compatible with Exasol - e.g. when \'final\' is being used which is a keyword in Exasol. Please visit [Adaopter dispatch documentation](https://docs.getdbt.com/guides/advanced/adapter-development/3-building-a-new-adapter#adapter-dispatch) of dbt-labs for more information. \n# Reporting bugs and contributing code\n- Please report bugs using the issues\n\n# Release\n## Release 1.2.0\n- support for invalidate_hard_deletes option in snapshots added by jups23\n- added persist_docs support by sti0\n- added additional configuration keys that can be included in profiles.yml by johannes-becker-otto\n- added cross-database macros introduced in 1.2 by sti0\n- added support for connection retries by sti0\n- added support for grants by sti0\n- added pytest functional adapter tests by tglunde\n- tox testing for python 3.7.2 through 3.10 added by tglunde\n \n## Release 1.0.0\n- pyexasol HTTP import csv feature implemented. Optimal performance and compatibility with Exasol CSV parsing\n',
    'author': 'Torsten Glunde',
    'author_email': 'torsten.glunde@alligator-company.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://alligatorcompany.gitlab.io/dbt-exasol',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



# pylint: disable=wrong-import-order
"""
DBT adapter connection implementation for Exasol.
"""
import decimal
import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, List, Optional

import agate
import dbt.exceptions
import pyexasol
from dbt.adapters.base import Credentials  # type: ignore
from dbt.adapters.exasol.relation import ProtocolVersionType
from dbt.adapters.sql import SQLConnectionManager  # type: ignore
from dbt.contracts.connection import AdapterResponse
from dbt.logger import GLOBAL_LOGGER as logger  # type: ignore
from pyexasol import ExaConnection


def connect(**kwargs: bool):
    """
    Global connect method initialising ExasolConnection
    """
    if "autocommit" not in kwargs:
        kwargs["autocommit"] = False

    return ExasolConnection(**kwargs)


class ExasolConnection(ExaConnection):
    """
    Override to instantiate ExasolCursor
    """

    def cursor(self):
        """Instance of ExasolCursor"""
        return ExasolCursor(self)


@dataclass
class ExasolAdapterResponse(AdapterResponse):
    """
    Override AdapterResponse
    """

    execution_time: Optional[float] = None


@dataclass
class ExasolCredentials(Credentials):
    """Profile parameters for Exasol in dbt profiles.yml"""

    dsn: str
    user: str
    password: str
    database: str
    schema: str
    # optional statements that can be set in profiles.yml
    # some options might interfere with dbt, so caution is advised
    connection_timeout: int = pyexasol.constant.DEFAULT_CONNECTION_TIMEOUT
    socket_timeout: int = pyexasol.constant.DEFAULT_SOCKET_TIMEOUT
    query_timeout: int = pyexasol.constant.DEFAULT_QUERY_TIMEOUT
    compression: bool = False
    encryption: bool = False
    ## Because of potential interference with dbt, the following statements are not (yet) implemented
    # fetch_dict: bool
    # fetch_size_bytes: int
    # lower_ident: bool
    # quote_ident: bool
    # verbose_error: bool
    # debug: bool
    # udf_output_port: int
    protocol_version: str = "v3"
    retries: int = 1

    _ALIASES = {"dbname": "database", "pass": "password"}

    @property
    def type(self):
        return "exasol"

    @property
    def unique_field(self):
        return self.dsn

    def _connection_keys(self):
        return (
            "dsn",
            "user",
            "database",
            "schema",
            "connection_timeout",
            "socket_timeout",
            "query_timeout",
            "compression",
            "encryption",
            "protocol_version",
        )


class ExasolConnectionManager(SQLConnectionManager):
    """Managing Exasol connections"""

    TYPE = "exasol"

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield

        except Exception as yielded_exception:
            logger.debug(f"Error running SQL: {sql}")
            logger.debug("Rolling back transaction.")
            self.release()
            if isinstance(yielded_exception, dbt.exceptions.RuntimeException):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise dbt.exceptions.RuntimeException(yielded_exception)

    @classmethod
    def get_result_from_cursor(cls, cursor: Any) -> agate.Table:
        data: List[Any] = []
        column_names: List[str] = []

        if cursor.description is not None:
            # column_names = [col[0] for col in cursor.description]
            rows = cursor.fetchall()
            for idx, col in enumerate(cursor.description):
                column_names.append(col[0])
                if len(rows) > 0 and isinstance(rows[0][idx], str):
                    if col[1] in ["DECIMAL", "BIGINT"]:
                        for rownum, row in enumerate(rows):
                            tmp = list(row)
                            tmp[idx] = decimal.Decimal(row[idx])
                            rows[rownum] = tmp
            data = cls.process_results(column_names, rows)

        return dbt.clients.agate_helper.table_from_data_flat(data, column_names)  # type: ignore

    @classmethod
    def open(cls, connection):
        if connection.state == "open":
            logger.debug("Connection is already open, skipping open.")
            return connection
        credentials = cls.get_credentials(connection.credentials)

        # Support protocol versions
        try:
            format_protocol_version = credentials.protocol_version.lower()
            version = ProtocolVersionType(format_protocol_version)

            if version == ProtocolVersionType.V1:
                protocol_version = pyexasol.PROTOCOL_V1
            elif version == ProtocolVersionType.V2:
                protocol_version = pyexasol.PROTOCOL_V2
            else:
                protocol_version = pyexasol.PROTOCOL_V3
        except:
            raise dbt.exceptions.RuntimeException(
                f"{credentials.protocol_version} is not a valid protocol version."
            )

        def _connect():
            return connect(
                dsn=credentials.dsn,
                user=credentials.user,
                password=credentials.password,
                autocommit=True,
                connection_timeout=credentials.connection_timeout,
                socket_timeout=credentials.socket_timeout,
                query_timeout=credentials.query_timeout,
                compression=credentials.compression,
                encryption=credentials.encryption,
                protocol_version=protocol_version,
            )

        retryable_exceptions = [pyexasol.ExaError]

        return cls.retry_connection(
            connection,
            connect=_connect,
            logger=logger,
            retry_limit=credentials.retries,
            retryable_exceptions=retryable_exceptions,
        )

    def commit(self):
        connection = self.get_thread_connection()
        if dbt.flags.STRICT_MODE:  # type: ignore
            assert isinstance(connection, ExaConnection)

        logger.debug(f"On {connection.name}: COMMIT")
        self.add_commit_query()

        connection.transaction_open = False

        return connection

    def begin(self):
        connection = self.get_thread_connection()

        if dbt.flags.STRICT_MODE:
            assert isinstance(connection, ExaConnection)

        if connection.transaction_open is True:
            raise dbt.exceptions.InternalException(
                f"Tried to begin a new transaction on connection {connection.name}, "
                "but it already had one open!"
            )

        connection.transaction_open = True
        return connection

    def cancel(self, connection):
        connection.abort_query()  # type: ignore

    @classmethod
    def get_status(cls, cursor):
        """Override status"""
        return "OK"

    @classmethod
    def get_response(cls, cursor) -> ExasolAdapterResponse:
        return ExasolAdapterResponse(
            _message="OK",
            rows_affected=cursor.rowcount,
            execution_time=cursor.execution_time,
        )

    @classmethod
    def get_credentials(cls, credentials):
        """override get_credentials"""
        return credentials

    def add_query(self, sql, auto_begin=True, bindings=None, abridge_sql_log=False):
        connection = self.get_thread_connection()
        if auto_begin and connection.transaction_open is False:
            self.begin()
        logger.debug(sql)
        logger.debug(f'Using {self.TYPE} connection "{connection.name}".')

        if sql.startswith("0CSV|"):
            connection.handle.cursor().import_from_file(bindings, sql.split("|", 1)[1])  # type: ignore

            return connection

        with self.exception_handler(sql):
            if abridge_sql_log:
                logger.debug("On {}: {}....".format(connection.name, sql[0:512]))
            else:
                logger.debug("On {}: {}".format(connection.name, sql))
            pre = time.time()

            cursor = connection.handle.cursor().execute(sql)  # type: ignore

            logger.debug(
                "SQL status: {} in {:.2f} seconds".format(
                    self.get_status(cursor), (time.time() - pre)
                )
            )

            return connection, cursor


class ExasolCursor(object):
    """Exasol dbt-adapter cursor implementation"""

    array_size = 1

    def __init__(self, connection):
        self.connection = connection
        self.stmt = None

    def import_from_file(self, agate_table, table):
        """importing csv skip=1 parameter for header row"""
        self.connection.import_from_file(
            agate_table.original_abspath,
            (table.split(".")[0], table.split(".")[1]),
            import_params={"skip": 1},
        )

    def execute(self, query):
        """executing query"""
        self.stmt = self.connection.execute(query)
        return self

    def executemany(self, query):
        """execute many not implemented yet"""
        raise RuntimeError

    def fetchone(self):
        return self.stmt.fetchone()

    def fetchmany(self, size=None):
        if size is None:
            size = self.array_size

        return self.stmt.fetchmany(size)

    def fetchall(self):
        return self.stmt.fetchall()

    def nextset(self):
        raise RuntimeError

    def setinputsizes(self):
        pass

    def setoutputsize(self):
        pass

    @property
    def description(self):
        cols = []
        if "resultSet" != self.stmt.result_type:
            return None

        for k, v in self.stmt.columns().items():
            cols.append(
                (
                    k,
                    v.get("type", None),
                    v.get("size", None),
                    v.get("size", None),
                    v.get("precision", None),
                    v.get("scale", None),
                    True,
                )
            )

        return cols

    @property
    def rowcount(self):
        return self.stmt.rowcount()

    @property
    def execution_time(self):
        return self.stmt.execution_time

    def close(self):
        self.stmt.close()
